<?php

/*
  Движок Типографа - http://rmcreative.ru/blog/post/tipograf
*/

class Typographus_Lite_UTF8{	
	private $_sym = array(
		'nbsp'    => '&nbsp;',
		'lnowrap' => '<span style="white-space:nowrap">',
		'rnowrap' => '</span>',

		'lquote'  => '«',
		'rquote'  => '»',
		'lquote2' => '„',
		'rquote2' => '“',
		'mdash'   => '—',
		'ndash'   => '–',
		'minus'   => '–', // соотв. по ширине символу +, есть во всех шрифтах

		'hellip'  => '…',
		'copy'    => '©',
		'trade'   => '™',
		'apos'    => '&#39;',   // см. http://fishbowl.pastiche.org/2003/07/01/the_curse_of_apos
		'reg'     => '®',
		'multiply' => '&times;',
		'1/2' => '&frac12;',
		'1/4' => '&frac14;',
		'3/4' => '&frac34;',
		'plusmn' => '&plusmn;',
		'rarr' => '&rarr;',
		'larr' => '&larr;',
	    'rsquo' => '&rsquo;'
	);

	private $_safeBlocks = array();

	/**
	 * Добавляет безопасный блок, который не будет обрабатываться типографом.
	 *
	 * @param string $openTag
	 * @param string $closeTag
	 */
	public function addSafeBlock($openTag, $closeTag){
		$this->_safeBlocks[$openTag] = $closeTag;
	}
	
	/**
	 * Убирает все безопасные блоки типографа.
	 * Полезно, если необходимо задать полностью свой набор.
	 */
	public function removeAllSafeBlocks(){
		$this->_safeBlocks = array();
	}

	/**
	 * Устанавливает соответствие между символом и его представлением.
	 *
	 * @param String $sym
	 * @param String $entity
	 */
	public function setSym($sym, $entity){
		$this->_sym[$sym] = $entity;
	}
	
	/**
	 * Вызывает типограф, обходя html-блоки и безопасные блоки
	 *
	 * @param string $str - обрабатываемый текст
	 * @param array $options - опции фильтрации
	 * @return string - обработанный текст
	 */
	public function process($str,$options){
		//создается шаблон поиска безопасных блоков
		$pattern = '(';
		foreach ($this->_safeBlocks as $start => $end){
			$pattern .= "$start.*$end|";
		}
		$pattern .= '<[^>]*[\s][^>]*>)';
		//блоки выдергиваются из текста, заменяясь на <номер блока>
		$str = preg_replace_callback("~$pattern~isU", array('self', '_stack'), $str);
		//собственно, обработка
		$str = $this->typo_text($str,$options);
		//возвращаются сейфблоки
		$str = strtr($str, self::_stack());
		return $str;
	}

	/**
	 * Накапливает исходный код безопасных блоков при использовании в качестве
	 * обратного вызова. При самостоятельном вызове без аргументоов - возвращает накопленный
	 * массив.
	 *
	 * @param array $matches
	 * @return array
	 */
	private static function _stack($matches = false){
		static $safe_blocks = array();
		if ($matches !== false){
			$key = '<'.count($safe_blocks).'>';
			$safe_blocks[$key] = $matches[0];
			return $key;
		}
		else{
			$tmp = $safe_blocks;
			unset($safe_blocks);
			return $tmp;
		}
	}
	
	/**
	 * Применяет все переданные правила к тексту
	 * @param array
	 * @return String
	 */
	private function apply_rules($rules, $str){
      return preg_replace(array_keys($rules), array_values($rules), $str);
	}

	/**
	 * Главная функция типографа.
	 * @param String $str - обрабатываемый текст
	 * @param Array $options - опции фильтрации
	 * @return String - исправленный текст
	 */
	private function typo_text($str,$options){
		$sym = $this->_sym;
		if (trim($str) == '') return '';

		$html_tag = '(?:(?U)<.*>)';
		$hellip = '\.{3,5}';
		
		//Слово
		$word = '[a-zA-Z_абвгдеёжзийклмнопрстуфхцчшщьыъэюяАБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЬЫЪЭЮЯ]';
		
		//Начало слова
		$phrase_begin = "(?:$hellip|$word|\n)";
		//Конец слова
        $phrase_end   = '(?:[)!?.:;#*\\\]|$|'.$word.'|'.$sym['rquote'].'|'.$sym['rquote2'].'|&quot;|"|'.$sym['hellip'].'|'.$sym['copy'].'|'.$sym['trade'].'|'.$sym['apos'].'|'.$sym['reg'].'|\')';
       
		$any_quote = "(?:$sym[lquote]|$sym[rquote]|$sym[lquote2]|$sym[rquote2]|&quot;|\")";

		//ликвидация кучи пробелов и табуляций
		$rules_strict = array();
		
		if (isset($options['fix_spaces'])) {
			// Много пробелов или табуляций -> один пробел
			$rules_strict['~( |\t)+~'] = ' ';
		}
		
		//автоподстановка
		$rules_symbols = array();
		
		if (isset($options['replace_copyright'])) {
			//(c)
			$rules_symbols['~\((c|с)\)~i'] = $sym['copy'];
		}
		if (isset($options['replace_registered'])) {
			//(r)
			$rules_symbols['~\(r\)~i'] = $sym['reg'];
		}
		if (isset($options['replace_trademark'])) {
			//tm
			$rules_symbols['~\(tm\)~i'] = $sym['trade'];
		}
		if (isset($options['replace_hellip'])) {
			//многоточие
			$rules_symbols["~$hellip~"] = $sym['hellip'];
		}
		if (isset($options['replace_plusminus'])) {
			//+-
			$rules_symbols['~([^\+]|^)\+-~'] = '$1'.$sym['plusmn'];
		}
		     
        //управление кавычками
        $rules_quotes = array();
        
        if (isset($options['fix_quotes'])) {
             // Разносим неправильные кавычки
		     $rules_quotes['~([^"]\w+)"(\w+)"~'] = '$1 "$2"';
		     $rules_quotes['~"(\w+)"(\w+)~'] = '"$1" $2';

             // Превращаем кавычки в ёлочки. Двойные кавычки склеиваем.
             $rules_quotes["~(?<=\\s|^|[>(])($html_tag*)($any_quote)($html_tag*$phrase_begin$html_tag*)~"] = '$1'.$sym['lquote'].'$3';
             $rules_quotes["~($html_tag*(?:$phrase_end|[0-9]+)$html_tag*)($any_quote)($html_tag*$phrase_end$html_tag*|\\s|[,<-])~"]	= '$1'.$sym['rquote'].'$3';
        }
            
		//куча основных правил
		$rules_main = array();
		
		if (isset($options['fix_dashes'])) {
			// Знак дефиса или два знака дефиса подряд — на знак длинного тире.
			// + Нельзя разрывать строку перед тире, например: Знание — сила, Курить — здоровью вредить.
			$rules_main['~ +(?:--?|—|&mdash;)(?=\s)~'] = $sym['nbsp'].$sym['mdash'];
		    $rules_main['~^(?:--?|—|&mdash;)(?=\s)~'] = $sym['mdash'];
    		//Прямая речь
		    $rules_main['~(?:^|\s+)(?:--?|—|&mdash;)(?=\s)~'] = "\n".$sym['nbsp'].$sym['mdash'];
		}
		if (isset($options['fix_ndash'])) {
			// Знак дефиса, ограниченный с обоих сторон цифрами — на знак короткого тире.
			$rules_main['~(?<=\d)-(?=\d)~'] = $sym['ndash'];
		}
		if (isset($options['glue_percent'])) {
			//% не отделяется от числа
		    $rules_main['~([0-9]+)\s+%~'] = '$1%';
		}
		if (isset($options['glue_IP'])) {
			//IP-адреса рвать нехорошо
			$rules_main['~(1\d{0,2}|2(\d|[0-5]\d)?)\.(0|1\d{0,2}|2(\d|[0-5]\d)?)\.(0|1\d{0,2}|2(\d|[0-5]\d)?)\.(0|1\d{0,2}|2(\d|[0-5]\d)?)~'] =	$sym['lnowrap'].'$0'.$sym['rnowrap'];
		}
		
		$str = $this->apply_rules($rules_quotes, $str);
				
		// Вложенные кавычки.
		if (isset($options['nested_quotes'])) {
	        $i=0; $lev = 5;
	        while (($i<$lev) && preg_match('~«(?:[^»]*?)«~', $str)){
				$i++;
	            $str = preg_replace('~«([^»]*?)«(.*?)»~s', '«$1'.$sym['lquote2'].'$2'.$sym['rquote2'], $str);
			}
	
	        $i=0;
	        while (($i++<$lev) && preg_match('~»(?:[^«]*?)»~', $str)){
			  $i++;
	          $str = preg_replace('~»([^«]*?)»~', $sym['rquote2'].'$1»', $str);
			}
		}
        
        $str = $this->apply_rules($rules_strict, $str);        
		$str = $this->apply_rules($rules_main, $str);
		$str = $this->apply_rules($rules_symbols, $str);
		
		return $str;
	}
}
